package org.nakedobjects.nos.store.file;

import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.persist.ObjectNotFoundException;
import org.nakedobjects.noa.persist.ObjectPersistenceException;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.persist.SerialOid;
import org.nakedobjects.nof.core.util.Assert;

import java.util.Enumeration;
import java.util.Vector;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;


public class XmlDataManager implements DataManager {
    private final XmlFile xmlFile;

    public XmlDataManager(XmlFile xmlFile) {
        this.xmlFile = xmlFile;
    }

    public String getDebugData() {
        return "Data directory " + XmlFile.directory();
    }

    // TODO the following methods are being called repeatedly - is there no
    // caching? See the print statements
    private class DataHandler extends DefaultHandler {
        CollectionData collection;
        StringBuffer data = new StringBuffer();
        String fieldName;
        ObjectData object;

        public void characters(final char[] ch, final int start, final int end) throws SAXException {
            data.append(new String(ch, start, end));
            //System.out.println("XML DataHandler " + data);
        }

        public void endElement(final String ns, final String name, final String tagName) throws SAXException {
            if (object != null) {
                if (tagName.equals("value")) {
                    String value = data.toString();
                    object.set(fieldName, value);
                    //System.out.println("XML DataHandler " + data);
                }
            }
        }

        public void startElement(final String ns, final String name, final String tagName, final Attributes attrs)
                throws SAXException {
            if (object != null) {
                if (tagName.equals("value")) {
                    fieldName = attrs.getValue("field");
                    data.setLength(0);
                    //System.out.println("XML DataHandler" + fieldName);
                } else if (tagName.equals("association")) {
                    String fieldName = attrs.getValue("field");
                    long id = Long.valueOf(attrs.getValue("ref"), 16).longValue();
                    object.set(fieldName, SerialOid.createPersistent(id));
                } else if (tagName.equals("element")) {
                    long id = Long.valueOf(attrs.getValue("ref"), 16).longValue();
                    object.addElement(fieldName, SerialOid.createPersistent(id));
                } else if (tagName.equals("multiple-association")) {
                    fieldName = attrs.getValue("field");
                    SerialOid internalCollection = null; // new
                    object.initCollection(internalCollection, fieldName);
                }
            } else if (collection != null) {
                if (tagName.equals("element")) {
                    long id = Long.valueOf(attrs.getValue("ref"), 16).longValue();
                    collection.addElement(SerialOid.createPersistent(id));
                }
            } else {
                if (tagName.equals("naked-object")) {
                    String type = attrs.getValue("type");
                    long version = Long.valueOf(attrs.getValue("ver"), 16).longValue();
                    String user = attrs.getValue("user");
                    long id = Long.valueOf(attrs.getValue("id"), 16).longValue();
                    NakedObjectSpecification spec = NakedObjectsContext.getReflector().loadSpecification(type);
                    SerialOid oid = SerialOid.createPersistent(id);
                    object = new ObjectData(spec, oid, new FileVersion(user, version));
                } else if (tagName.equals("collection")) {
                    String type = attrs.getValue("type");
                    long version = Long.valueOf(attrs.getValue("ver"), 16).longValue();
                    String user = attrs.getValue("user");
                    long id = Long.valueOf(attrs.getValue("id"), 16).longValue();
                    NakedObjectSpecification spec = NakedObjectsContext.getReflector().loadSpecification(type);
                    SerialOid oid = SerialOid.createPersistent(id);
                    collection = new CollectionData(spec, oid, new FileVersion(user, version));
                } else {
                    throw new SAXException("Invalid data");
                }
            }
        }
    }

    private class InstanceHandler extends DefaultHandler {
        Vector instances = new Vector();

        public void characters(final char[] arg0, final int arg1, final int arg2) throws SAXException {}

        public void startElement(final String ns, final String name, final String tagName, final Attributes attrs)
                throws SAXException {
            if (tagName.equals("instance")) {
                long oid = Long.valueOf(attrs.getValue("id"), 16).longValue();
                instances.addElement(SerialOid.createPersistent(oid));
            }
        }
    }

    private class NumberHandler extends DefaultHandler {
        boolean captureValue = false;
        long value = 0;

        public void characters(final char[] arg0, final int arg1, final int arg2) throws SAXException {
            if (captureValue) {
                value = Long.valueOf(new String(arg0, arg1, arg2), 16).longValue();
            }
        }

        public void startElement(final String ns, final String name, final String tagName, final Attributes attrs)
                throws SAXException {
            captureValue = tagName.equals("number");
        }
    }

    /**
     * Write out the data for a new instance.
     */
    protected void addData(final SerialOid oid, final String type, final Data data) {
        writeData(oid, data);
    }

    /**
     * Add the reference for an instance to the list of all instances.
     */
    protected void addInstance(final SerialOid oid, final String type) {
        Vector instances = loadInstances(type);
        instances.addElement(oid);
        writeInstanceFile(type, instances);
    }

    private String attribute(final String name, final String value) {
        return " " + name + "=\"" + value + "\"";
    }

    public final SerialOid createOid() throws PersistorException {
        return SerialOid.createPersistent(nextId());
    }

    /**
     * Delete the data for an existing instance.
     */
    protected void deleteData(final SerialOid oid, final String type) {
        xmlFile.delete(filename(oid));
    }

    private String encodedOid(final SerialOid oid) {
        return Long.toHexString(oid.getSerialNo()).toUpperCase();
    }

    private String filename(final SerialOid oid) {
        return encodedOid(oid);
    }

    public ObjectDataVector getInstances(final ObjectData pattern) {
        Vector instances = loadInstances(pattern.getClassName());

        if (instances == null) {
            return new ObjectDataVector();
        }

        ObjectDataVector matches = new ObjectDataVector();
        for (int i = 0; i < instances.size(); i++) {
            SerialOid oid = (SerialOid) instances.elementAt(i);
            ObjectData instanceData = (ObjectData) loadData(oid);
            // TODO check loader first
            if (instanceData == null) {
                throw new NakedObjectRuntimeException("No data found for " + oid + " (possible missing file)");
            }
            if (matchesPattern(pattern, instanceData)) {
                matches.addElement(instanceData);
            }
        }
        return matches;
    }

    public void getNakedClass(final String name) {}

    /**
     * Save the data for an object and adds the reference to a list of instances
     */
    public final void insertObject(final ObjectData data) {
        if (data.getOid() == null) {
            throw new IllegalArgumentException("Oid must be non-null");
        }

        String type = data.getClassName();
        SerialOid oid = data.getOid();
        addData(oid, type, data);
        addInstance(oid, type);
    }

    public boolean isInitialized() {
        return xmlFile.isInitialized();
    }
    
    /**
     * Loads in data for a collection for the specified identifier.
     */
    public final CollectionData loadCollectionData(SerialOid oid) {
        return (CollectionData) loadData(oid);
    }

    public Data loadData(final SerialOid oid) {
        DataHandler handler = new DataHandler();
        xmlFile.parse(handler, filename(oid));

        if (handler.object != null) {
            return handler.object;
        } else {
            return handler.collection;
        }
    }

    private Vector loadInstances(final String type) {
        InstanceHandler handler = new InstanceHandler();
        xmlFile.parse(handler, type);
        return handler.instances;
    }

    /**
     * Loads in data for an object for the specified identifier.
     */
    public final ObjectData loadObjectData(SerialOid oid) {
        return (ObjectData) loadData(oid);
    }

    /**
     * A helper that determines if two sets of data match. A match occurs when the types are the same and any
     * field in the pattern also occurs in the data set under test.
     */
    // TODO we need to be able to find collection instances as well
    protected boolean matchesPattern(final ObjectData patternData, final ObjectData testData) {
        if (patternData == null || testData == null) {
            throw new NullPointerException("Can't match on nulls " + patternData + " " + testData);
        }
        if (!patternData.getClassName().equals(testData.getClassName())) {
            return false;
        }

        Enumeration fields = patternData.fields();

        while (fields.hasMoreElements()) {
            String field = (String) fields.nextElement();
            Object patternFieldValue = patternData.get(field);

            Object testFieldValue = testData.get(field);

            if (testFieldValue instanceof ReferenceVector) {
                ReferenceVector patternElements = (ReferenceVector) patternFieldValue;
                for (int i = 0; i < patternElements.size(); i++) {
                    SerialOid requiredElement = patternElements.elementAt(i); // must
                    // have
                    // this
                    // element
                    boolean requiredFound = false;
                    ReferenceVector testElements = ((ReferenceVector) testFieldValue);
                    for (int j = 0; j < testElements.size(); j++) {
                        if (requiredElement.equals(testElements.elementAt(j))) {
                            requiredFound = true;
                            break;
                        }
                    }
                    if (!requiredFound) {
                        return false;
                    }
                }
            } else {
                if (!patternFieldValue.equals(testFieldValue)) {
                    return false;
                }
            }

        }

        return true;
    }

    /**
     * Read in the next unique number for the object identifier.
     */
    protected long nextId() throws PersistorException {
        NumberHandler handler = new NumberHandler();
        xmlFile.parse(handler, "oid");

        StringBuffer data = new StringBuffer();
        data.append("<number>");
        data.append(handler.value + 1);
        data.append("</number>");
        xmlFile.writeXml("oid", data);

        return handler.value + 1;
    }

    public int numberOfInstances(final ObjectData pattern) {
        Vector instances = loadInstances(pattern.getClassName());

        if (instances == null) {
            return 0;
        }

        int instanceCount = 0;
        for (int i = 0; i < instances.size(); i++) {
            SerialOid oid = (SerialOid) instances.elementAt(i);
            ObjectData instanceData = (ObjectData) loadData(oid);
            if (instanceData != null && matchesPattern(pattern, instanceData)) {
                instanceCount++;
            }
        }
        return instanceCount;
    }

    public final void remove(SerialOid oid) throws ObjectNotFoundException, ObjectPersistenceException {
        Data data = loadData(oid);
        String type = data.getClassName();
        removeInstance(oid, type);
        deleteData(oid, type);
    }

    /**
     * Remove the reference for an instance from the list of all instances.
     */
    protected void removeInstance(final SerialOid oid, final String type) {
        Vector instances = loadInstances(type);
        instances.removeElement(oid);
        writeInstanceFile(type, instances);
    }

    /**
     * Save the data for latter retrieval.
     */
    public final void save(Data data) {
        writeData(data.getOid(), data);
    }

    public void shutdown() {}

    private void writeData(final SerialOid xoid, final Data data) {
        StringBuffer xml = new StringBuffer();
        boolean isObject = data instanceof ObjectData;
        String tag = isObject ? "naked-object" : "collection";
        xml.append("<" + tag);
        xml.append(attribute("type", data.getClassName()));
        xml.append(attribute("id", "" + encodedOid(data.getOid())));
        xml.append(attribute("user", "" + NakedObjectsContext.getSession().getUserName()));

        long sequence = data.getVersion().getSequence();
        String sequenceString = Long.toHexString(sequence).toUpperCase();
        xml.append(attribute("ver", "" + sequenceString));

        xml.append(">\n");

        if (isObject) {
            ObjectData object = (ObjectData) data;
            Enumeration fields = object.fields();

            while (fields.hasMoreElements()) {
                String field = (String) fields.nextElement();
                Object entry = object.get(field);

                if (entry instanceof SerialOid) {
                    Assert.assertFalse(((SerialOid) entry).isTransient());
                    xml.append("  <association field=\"" + field + "\" ");
                    xml.append("ref=\"" + encodedOid((SerialOid) entry) + "\"/>\n");
                } else if (entry instanceof ReferenceVector) {
                    ReferenceVector references = (ReferenceVector) entry;
                    int size = references.size();

                    if (size > 0) {
                        xml.append("  <multiple-association field=\"" + field + "\" ");
                        xml.append(">\n");
                        for (int i = 0; i < size; i++) {
                            Object oid = references.elementAt(i);
                            Assert.assertFalse("transient oid", oid, ((SerialOid) oid).isTransient());
                            xml.append("    <element ");
                            xml.append("ref=\"" + encodedOid((SerialOid) oid) + "\"/>\n");
                        }
                        xml.append("  </multiple-association>\n");
                    }
                } else {
                    xml.append("  <value field=\"" + field + "\">");
                    xml.append(XmlFile.getValueWithSpecialsEscaped(entry.toString()));
                    xml.append("</value>\n");
                }
            }
        } else {
            CollectionData collection = (CollectionData) data;
            ReferenceVector refs = collection.references();
            for (int i = 0; i < refs.size(); i++) {
                Object oid = refs.elementAt(i);
                xml.append("  <element ");
                xml.append("ref=\"" + encodedOid((SerialOid) oid) + "\"/>\n");
            }
        }

        xml.append("</" + tag + ">\n");
        xmlFile.writeXml(filename(xoid), xml);
    }

    private void writeInstanceFile(final String name, final Vector instances) {
        StringBuffer data = new StringBuffer();
        data.append("<instances name=\"" + name + "\">\n");

        for (int i = 0; i < instances.size(); i++) {
            data.append("  <instance id=\"" + encodedOid((SerialOid) instances.elementAt(i)) + "\"/>\n");
        }

        data.append("</instances>");
        xmlFile.writeXml(name, data);
    }

}
// Copyright (c) Naked Objects Group Ltd.
