package org.nakedobjects.nos.store.file;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.Naked;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedObjectLoader;
import org.nakedobjects.noa.adapter.NakedValue;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.adapter.ResolveState;
import org.nakedobjects.noa.persist.InstancesCriteria;
import org.nakedobjects.noa.persist.ObjectNotFoundException;
import org.nakedobjects.noa.persist.ObjectPersistenceException;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.reflect.OneToManyAssociation;
import org.nakedobjects.noa.reflect.OneToOneAssociation;
import org.nakedobjects.noa.reflect.ValueAssociation;
import org.nakedobjects.noa.spec.Features;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.persist.SerialOid;
import org.nakedobjects.nof.core.util.Assert;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.persist.objectstore.NakedObjectStore;
import org.nakedobjects.nof.persist.transaction.CreateObjectCommand;
import org.nakedobjects.nof.persist.transaction.DestroyObjectCommand;
import org.nakedobjects.nof.persist.transaction.ExecutionContext;
import org.nakedobjects.nof.persist.transaction.PersistenceCommand;
import org.nakedobjects.nof.persist.transaction.SaveObjectCommand;


public class XmlObjectStore implements NakedObjectStore {
    private static final Logger LOG = Logger.getLogger(XmlObjectStore.class);
    private DataManager dataManager;
    private ServiceManager serviceManager;
    private boolean isInitialized;

    public XmlObjectStore() {
        XmlFile xmlFile = new XmlFile();
        dataManager = new XmlDataManager(xmlFile);
        serviceManager = new XmlServiceManager(xmlFile);
        serviceManager.loadServices();
    }

    public XmlObjectStore(DataManager dataManager, ServiceManager serviceManager) {
        this.dataManager = dataManager;
        this.serviceManager = serviceManager;
        serviceManager.loadServices();
    }

    public void abortTransaction() {
        LOG.debug("transaction aborted");
    }

    private NakedObjectSpecification specFor(final Data data) {
        return NakedObjectsContext.getReflector().loadSpecification(data.getClassName());
    }

    public CreateObjectCommand createCreateObjectCommand(final NakedObject object) {
        return new CreateObjectCommand() {
            public void execute(final ExecutionContext context) throws ObjectPersistenceException {
                LOG.debug("  create object " + object);
                String user = NakedObjectsContext.getSession().getUserName();
                object.setOptimisticLock(new FileVersion(user));
                ObjectData data = createObjectData(object, true);
                dataManager.insertObject(data);
            }

            public NakedObject onObject() {
                return object;
            }

            public String toString() {
                return "CreateObjectCommand [object=" + object + "]";
            }
        };
    }

    public void registerService(final String name, final Oid oid) {
        serviceManager.registerService(name, oid);
    }

    public DestroyObjectCommand createDestroyObjectCommand(final NakedObject object) {
        return new DestroyObjectCommand() {
            public void execute(final ExecutionContext context) throws ObjectPersistenceException {
                LOG.debug("  destroy object " + object);
                SerialOid oid = (SerialOid) object.getOid();
                dataManager.remove(oid);
                object.setOptimisticLock(null);
            }

            public NakedObject onObject() {
                return object;
            }

            public String toString() {
                return "DestroyObjectCommand [object=" + object + "]";
            }
        };
    }

    private ObjectData createObjectData(final NakedObject object, final boolean ensurePersistent) {
        LOG.debug("compiling object data for " + object);

        ObjectData data;
        data = new ObjectData(object.getSpecification(), (SerialOid) object.getOid(), (FileVersion) object.getVersion());

        NakedObjectField[] fields = object.getSpecification().getFields();

        for (int i = 0; i < fields.length; i++) {
            if (!fields[i].isPersisted()) {
                continue;
            }

            Naked field = fields[i].get(object);
            Naked fieldContent = field;
            String fieldId = fields[i].getId();

            if (fields[i].isCollection()) {
                data.addInternalCollection((NakedCollection) fieldContent, fieldId, ensurePersistent);
            } else if (fields[i].isValue()) {
                boolean isEmpty = fields[i].isEmpty(object);
                String encodedValue = ((NakedValue) field).asEncodedString();
                data.saveValue(fieldId, isEmpty, field == null || isEmpty ? null : new String(encodedValue));
            } else if (fields[i].isObject()) {
                data.addAssociation((NakedObject) fieldContent, fieldId, ensurePersistent);
            }
        }

        return data;
    }

    public SaveObjectCommand createSaveObjectCommand(final NakedObject object) {
        return new SaveObjectCommand() {
            public void execute(final ExecutionContext context) throws ObjectPersistenceException {
                LOG.debug("  save object " + object);

                if (Features.isAggregated(object.getSpecification())) {
                    NakedObject parent = object.getSpecification().getAggregate(object);
                    Data data = createObjectData(parent, true);
                    dataManager.save(data);
                } else {
                    String user = NakedObjectsContext.getSession().getUserName();
                    object.setOptimisticLock(new FileVersion(user));

                    Data data = createObjectData(object, true);
                    dataManager.save(data);
                }
            }

            public NakedObject onObject() {
                return object;
            }

            public String toString() {
                return "SaveObjectCommand [object=" + object + "]";
            }

        };
    }

    public void endTransaction() {
        LOG.debug("end transaction");
    }

    public void debugData(final DebugString debug) {
        debug.appendTitle("Business Objects");
        debug.appendln(dataManager.getDebugData());
    }

    public String debugTitle() {
        return "XML Object Store";
    }

    public NakedObject[] getInstances(final InstancesCriteria criteria) {
        LOG.debug("getInstances of " + criteria.getSpecification() + " where " + criteria);
        ObjectData patternData = new ObjectData(criteria.getSpecification(), null, null);
        NakedObject[] instances = getInstances(patternData, criteria);
        return instances;
    }

    private NakedObject[] getInstances(final ObjectData patternData, final InstancesCriteria criteria) {
        ObjectDataVector data = dataManager.getInstances(patternData);
        NakedObject[] instances = new NakedObject[data.size()];
        int count = 0;

        for (int i = 0; i < data.size(); i++) {
            ObjectData instanceData = data.element(i);
            LOG.debug("instance data " + instanceData);

            SerialOid oid = instanceData.getOid();

            NakedObjectSpecification spec = specFor(instanceData);
            NakedObject instance = objectLoader().recreateAdapterForPersistent(oid, spec);
            initObject(instance, instanceData);

            if (criteria == null || criteria.matches(instance)) {
                instances[count++] = instance;
            }
        }

        NakedObject[] array = new NakedObject[count];
        System.arraycopy(instances, 0, array, 0, count);
        return array;
    }

    private NakedObjectLoader objectLoader() {
        return NakedObjectsContext.getObjectLoader();
    }

    public NakedObject getObject(final Oid oid, final NakedObjectSpecification hint) {
        LOG.debug("getObject " + oid);
        Data data = dataManager.loadData((SerialOid) oid);
        LOG.debug("  data read " + data);

        NakedObject object;

        if (data instanceof ObjectData) {
            object = recreateObject((ObjectData) data);
        } else if (data instanceof CollectionData) {
            throw new NakedObjectRuntimeException();
        } else {
            throw new ObjectNotFoundException(oid);
        }
        return object;
    }

    public Oid getOidForService(String name) {
        return serviceManager.getOidForService(name);
    }

    public boolean hasInstances(final NakedObjectSpecification cls, final boolean includeSubclasses) {
        LOG.debug("checking instance of " + cls);
        ObjectData data = new ObjectData(cls, null, null);
        return dataManager.numberOfInstances(data) > 0;
    }

    public void init() throws ObjectPersistenceException {
        isInitialized = dataManager.isInitialized();
    }

    private void initObject(final NakedObject object, final ObjectData data) {
        if (object.getResolveState().isResolvable(ResolveState.RESOLVING)) {
            objectLoader().start(object, ResolveState.RESOLVING);

            NakedObjectField[] fields = object.getSpecification().getFields();
            for (int i = 0; i < fields.length; i++) {
                NakedObjectField field = fields[i];
                if (!field.isPersisted()) {
                    continue;
                }

                if (field.isValue()) {
                    NakedValue value = (NakedValue) ((ValueAssociation) field).get(object);
                    String valueData = data.value(field.getId());
                    value.restoreFromEncodedString(valueData == null ? "NULL" : valueData);
                    ((ValueAssociation) field).initValue(object, value);
                } else if (field.isCollection()) {
                    initObjectSetupCollection(object, data, field);
                } else if (field.isObject()) {
                    initObjectSetupReference(object, data, field);
                }
            }
            object.setOptimisticLock(data.getVersion());
            objectLoader().end(object);
        }
    }

    private void initObjectSetupReference(final NakedObject object, final ObjectData data, final NakedObjectField field) {
        SerialOid referenceOid = (SerialOid) data.get(field.getId());
        LOG.debug("setting up field " + field + " with " + referenceOid);
        if (referenceOid == null) {
            return;
        }

        Data fieldData = (Data) dataManager.loadData(referenceOid);

        if (fieldData == null) {
            NakedObject adapter = NakedObjectsContext.getObjectLoader().recreateAdapterForPersistent(referenceOid, field.getSpecification());
            if (!adapter.getResolveState().isDestroyed()) {
                adapter.changeState(ResolveState.DESTROYED);
            }
            ((OneToOneAssociation) field).initAssociation(object, adapter);
            
            LOG.warn("No data found for " + referenceOid + " so field '" + field.getName() + "' not set in object '"
                    + object.titleString() + "'");
        } else {
            NakedObject reference = objectLoader().recreateAdapterForPersistent(referenceOid, specFor(fieldData));
            ((OneToOneAssociation) field).initAssociation(object, reference);
        }

        /*
         * if (loadedObjects().isLoaded(referenceOid)) { NakedObject loadedObject =
         * loadedObjects().getLoadedObject(referenceOid); LOG.debug("using loaded object " + loadedObject);
         * object.initAssociation((OneToOneAssociation) field, loadedObject); } else { NakedObject
         * fieldObject; Data fieldData = (Data) dataManager.loadData((SerialOid) referenceOid);
         * 
         * if (fieldData != null) { fieldObject = (NakedObject) specFor(fieldData).acquireInstance(); } else {
         * fieldObject = (NakedObject) field.getSpecification().acquireInstance(); }
         * 
         * fieldObject.setOid(referenceOid);
         * 
         * if (fieldObject instanceof NakedCollection) { fieldObject.setResolved(); }
         * 
         * loadedObjects().loaded(fieldObject); object.initAssociation((OneToOneAssociation) field,
         * fieldObject); }
         */
    }

    private void initObjectSetupCollection(final NakedObject object, final ObjectData data, final NakedObjectField field) {
        /*
         * The internal collection is already a part of the object, and therefore cannot be recreated, but its
         * oid must be set
         */
        ReferenceVector refs = (ReferenceVector) data.get(field.getId());
        NakedCollection collection = (NakedCollection) field.get(object);
        NakedObjectsContext.getObjectLoader().start(collection, ResolveState.RESOLVING);
        int size = refs == null ? 0 : refs.size();
        NakedObject[] elements = new NakedObject[size];
        for (int j = 0; j < size; j++) {
            SerialOid elementOid = refs.elementAt(j);
            NakedObject adapter;
            if (objectLoader().isIdentityKnown(elementOid)) {
                adapter = objectLoader().getAdapterFor(elementOid);
            } else {
                adapter = getObject(elementOid, null);
            }
            elements[j] = adapter;
        }
        ((OneToManyAssociation) field).initCollection(object, elements);
        NakedObjectsContext.getObjectLoader().end(collection);
    }

    public boolean isInitialized() {
        return isInitialized;
    }

    public String name() {
        return "XML";
    }

    /*
     * The ObjectData holds all references for internal collections, so the object should haves its internal
     * collection populated by this method.
     */
    private NakedObject recreateObject(final ObjectData data) {
        SerialOid oid = data.getOid();
        NakedObjectSpecification spec = specFor(data);
        NakedObject object = objectLoader().recreateAdapterForPersistent(oid, spec);
        initObject(object, data);
        return object;
    }

    public void resolveField(final NakedObject object, final NakedObjectField field) {
        NakedObject reference = (NakedObject) field.get(object);
        resolveImmediately(reference);
    }

    public void resolveImmediately(final NakedObject object) {
        ObjectData data = (ObjectData) dataManager.loadData((SerialOid) object.getOid());
        Assert.assertNotNull("Not able to read in data during resolve", object, data);
        initObject(object, data);
    }

    public void reset() {}

    public void execute(final PersistenceCommand[] commands) {
        LOG.debug("start execution of transaction");
        for (int i = 0; i < commands.length; i++) {
            PersistenceCommand command = commands[i];
            command.execute(null);
        }
        LOG.debug("end execution");
    }

    public boolean flush(final PersistenceCommand[] commands) {
        LOG.debug("start flush of transaction");
        for (int i = 0; i < commands.length; i++) {
            PersistenceCommand command = commands[i];
            command.execute(null);
        }
        LOG.debug("end flush");
        return commands.length > 0;
    }

    /**
     * Set the clock used to generate sequence numbers and last changed dates for version objects.
     */
    public void setClock(final Clock clock) {
        FileVersion.setClock(clock);
    }

    /**
     * Set the clock used to generate sequence numbers and last changed dates for version objects.
     * 
     * @property
     */
    public void set_Clock(final Clock clock) {
        FileVersion.setClock(clock);
    }

    public void shutdown() {
        LOG.info("shutdown " + this);
    }

    public void startTransaction() {
        LOG.debug("start transaction");
    }
}
// Copyright (c) Naked Objects Group Ltd.
